@@ -11,11 +11,12 @@ import binascii
 import struct
 
 from intelhex import IntelHex
+
 from pynitrokey import helpers
 
 
 def genkey(output_pem_file, input_seed_file=None):
-    from ecdsa import SigningKey, NIST256p
+    from ecdsa import NIST256p, SigningKey
     from ecdsa.util import randrange_from_seed__trytryagain
 
     if input_seed_file is not None:
@@ -97,11 +98,11 @@ def mergehex(
 
     if attestation_key is None:
         # generic / hacker attestation key
-        print('*** Using development attestation key')
+        print("*** Using development attestation key")
         attestation_key = (
             "1b2626ecc8f69b0f69e34fb236d76466ba12ac16c3ab5750ba064e8b90e02448"
         )
-    assert(len(attestation_key) == 2*32)
+    assert len(attestation_key) == 2 * 32
 
     if attestation_cert is None:
         attestation_cert = hacker_attestation_cert
@@ -145,7 +146,7 @@ def mergehex(
     first[AUTH_WORD_ADDR + 7] = 0xFF
 
     # patch in the attestation key
-    print(f'Using attestation key[:2]: {attestation_key[:4]}...')
+    print(f"Using attestation key[:2]: {attestation_key[:4]}...")
     key = binascii.unhexlify(attestation_key)
 
     for i, x in enumerate(key):
@@ -154,7 +155,7 @@ def mergehex(
     offset = 32
 
     # patch in device settings / i.e. lock byte in little endian 64 int.
-    print(f'Setting lock = {lock}')
+    print(f"Setting lock = {lock}")
     lock_byte = 0x02 if lock else 0x00
     device_settings = struct.pack("<Q", 0xAA551E7900000000 | lock_byte)
 
@@ -178,7 +179,7 @@ def mergehex(
     first.tofile(output_hex_file, format="hex")
 
 
-def sign_firmware(sk_name, hex_file, APPLICATION_END_PAGE = 20, PAGES=128):
+def sign_firmware(sk_name, hex_file, APPLICATION_END_PAGE=20, PAGES=128):
     v1 = sign_firmware_for_version(sk_name, hex_file, 19)
     v2 = sign_firmware_for_version(sk_name, hex_file, 20, PAGES=PAGES)
 
@@ -201,10 +202,10 @@ def sign_firmware_for_version(sk_name, hex_file, APPLICATION_END_PAGE, PAGES=128
     # Maybe this is not the optimal module...
 
     import base64
-
     import binascii
-    from ecdsa import SigningKey
     from hashlib import sha256
+
+    from ecdsa import SigningKey
     from intelhex import IntelHex
 
     sk = SigningKey.from_pem(open(sk_name).read())
