@@ -8,18 +8,90 @@ import dataclasses
 import hmac
 import logging
 import typing
-from enum import Enum
+from enum import Enum, IntEnum
 from hashlib import pbkdf2_hmac
 from secrets import token_bytes
 from struct import pack
 from typing import List, Optional
 
+import semver
 import tlv8
 
 from pynitrokey.nk3 import Nitrokey3Device
 from pynitrokey.start.gnuk_token import iso7816_compose
 
 
+@dataclasses.dataclass
+class ListItemProperties:
+    touch_required: bool
+    secret_encryption: bool
+    pws_data_exist: bool
+
+    @classmethod
+    def _get_bit(cls, x: int, n: int) -> bool:
+        return ((x >> n) & 1) == 1
+
+    @classmethod
+    def from_byte(cls, b: int) -> "ListItemProperties":
+        return ListItemProperties(
+            touch_required=cls._get_bit(b, 0),
+            secret_encryption=cls._get_bit(b, 1),
+            pws_data_exist=cls._get_bit(b, 2),
+        )
+
+    def __str__(self) -> str:
+        data = [
+            "touch required" if self.touch_required else "",
+            "PIN required" if self.secret_encryption else "",
+            "PWS data available" if self.pws_data_exist else "",
+        ]
+        return ",".join([d for d in data if d])
+
+
+@dataclasses.dataclass
+class ListItem:
+    kind: "Kind"
+    algorithm: "Algorithm"
+    label: bytes
+    properties: ListItemProperties
+
+    @classmethod
+    def get_type_name(cls, x: typing.Any) -> str:
+        return str(x).split(".")[-1]
+
+    def __str__(self) -> str:
+        return (
+            f"{self.label.decode()}"
+            f"\t{self.get_type_name(self.kind)}/{self.get_type_name(self.algorithm)}"
+            f"\t{self.properties}"
+        )
+
+
+@dataclasses.dataclass
+class PasswordSafeEntry:
+    login: Optional[bytes]
+    password: Optional[bytes]
+    metadata: Optional[bytes]
+    properties: Optional[bytes] = None
+    name: Optional[bytes] = None
+
+    def tlv_encode(self) -> List[tlv8.Entry]:
+        entries = [
+            tlv8.Entry(Tag.PwsLogin.value, self.login)
+            if self.login is not None
+            else None,
+            tlv8.Entry(Tag.PwsPassword.value, self.password)
+            if self.password is not None
+            else None,
+            tlv8.Entry(Tag.PwsMetadata.value, self.metadata)
+            if self.metadata is not None
+            else None,
+        ]
+        # Filter out empty entries
+        entries = [r for r in entries if r is not None]
+        return entries
+
+
 @dataclasses.dataclass
 class RawBytes:
     data: List
@@ -37,6 +109,8 @@ class SelectResponse:
     challenge: Optional[bytes]
     # Selected algorithm, challenge-response auth only
     algorithm: Optional[bytes]
+    # Serial number of the device
+    serial_number: Optional[bytes]
 
     def version_str(self) -> str:
         if self.version:
@@ -48,15 +122,47 @@ class SelectResponse:
         return (
             "Nitrokey Secrets\n"
             f"\tVersion: {self.version_str()}\n"
-            f"\tPIN attempt counter: {self.pin_attempt_counter}"
+            f"\tPIN attempt counter: {self.pin_attempt_counter}\n"
+            f"\tSerial number: {self.serial_number.hex() if self.serial_number else 'None'}"
         )
 
 
+class SecretsAppExceptionID(IntEnum):
+    MoreDataAvailable = 0x61FF
+    VerificationFailed = 0x6300
+    UnspecifiedNonpersistentExecutionError = 0x6400
+    UnspecifiedPersistentExecutionError = 0x6500
+    WrongLength = 0x6700
+    LogicalChannelNotSupported = 0x6881
+    SecureMessagingNotSupported = 0x6882
+    CommandChainingNotSupported = 0x6884
+    SecurityStatusNotSatisfied = 0x6982
+    ConditionsOfUseNotSatisfied = 0x6985
+    OperationBlocked = 0x6983
+    IncorrectDataParameter = 0x6A80
+    FunctionNotSupported = 0x6A81
+    NotFound = 0x6A82
+    NotEnoughMemory = 0x6A84
+    IncorrectP1OrP2Parameter = 0x6A86
+    KeyReferenceNotFound = 0x6A88
+    InstructionNotSupportedOrInvalid = 0x6D00
+    ClassNotSupported = 0x6E00
+    UnspecifiedCheckingError = 0x6F00
+    Success = 0x9000
+
+
+class SecretsAppHealthCheckException(Exception):
+    pass
+
+
 @dataclasses.dataclass
 class SecretsAppException(Exception):
     code: str
     context: str
 
+    def to_id(self) -> SecretsAppExceptionID:
+        return SecretsAppExceptionID(int(self.code, 16))
+
     def to_string(self) -> str:
         d = {
             "61FF": "MoreDataAvailable",
@@ -84,12 +190,16 @@ class SecretsAppException(Exception):
         return d.get(self.code, "Unknown SW code")
 
     def __repr__(self) -> str:
-        return f"OTPAppException(code={self.code}/{self.to_string()})"
+        return f"SecretsAppException(code={self.code}/{self.to_string()})"
 
     def __str__(self) -> str:
         return self.__repr__()
 
 
+class CCIDInstruction(Enum):
+    Select = 0xA4
+
+
 class Instruction(Enum):
     Put = 0x1
     Delete = 0x2
@@ -98,14 +208,14 @@ class Instruction(Enum):
     List = 0xA1
     Calculate = 0xA2
     Validate = 0xA3
-    CalculateAll = 0xA4
+    CalculateAll = 0xA4  # 0xA4 is Select as well # Unused
     SendRemaining = 0xA5
     VerifyCode = 0xB1
-    Select = 0xA4
     # Place extending commands in 0xBx space
     VerifyPIN = 0xB2
     ChangePIN = 0xB3
     SetPIN = 0xB4
+    GetCredential = 0xB5
 
 
 class Tag(Enum):
@@ -123,18 +233,42 @@ class Tag(Enum):
     Password = 0x80
     NewPassword = 0x81
     PINCounter = 0x82
+    PwsLogin = 0x83
+    PwsPassword = 0x84
+    PwsMetadata = 0x85
+    SerialNumber = 0x8F
 
 
 class Kind(Enum):
     Hotp = 0x10
     Totp = 0x20
     HotpReverse = 0x30
+    Hmac = 0x40
+    NotSet = 0xF0
+
+    @classmethod
+    def from_attribute_byte(cls, attribute_byte: bytes) -> str:
+        a = int(attribute_byte)
+        k = cls.from_attribute_byte_type(a)
+        if k != Kind.NotSet:
+            return str(k).split(".")[-1].upper()
+        else:
+            return "PWS"
+
+    @classmethod
+    def from_attribute_byte_type(cls, a: int) -> "Kind":
+        v = a & 0xF0
+        for k in Kind:
+            if k.value == v:
+                return k
+        raise ValueError("Invalid attribute byte")
 
 
 STRING_TO_KIND = {
     "HOTP": Kind.Hotp,
     "TOTP": Kind.Totp,
     "HOTP_REVERSE": Kind.HotpReverse,
+    "HMAC": Kind.Hmac,
 }
 
 
@@ -144,6 +278,12 @@ class Algorithm(Enum):
     Sha512 = 0x03
 
 
+ALGORITHM_TO_KIND = {
+    "SHA1": Algorithm.Sha1,
+    "SHA256": Algorithm.Sha256,
+}
+
+
 class SecretsApp:
     """
     This is a Secrets App client
@@ -155,6 +295,7 @@ class SecretsApp:
     dev: Nitrokey3Device
     write_corpus_fn: Optional[typing.Callable]
     _cache_status: Optional[SelectResponse]
+    _metadata: dict
 
     def __init__(self, dev: Nitrokey3Device, logfn: Optional[typing.Callable] = None):
         self._cache_status = None
@@ -165,6 +306,7 @@ class SecretsApp:
         else:
             self.logfn = self.log.info  # type: ignore [assignment]
         self.dev = dev
+        self._metadata = {}
 
     def _custom_encode(self, structure: Optional[List] = None) -> bytes:
         if not structure:
@@ -174,17 +316,22 @@ class SecretsApp:
             if not d:
                 return b""
             if isinstance(d, RawBytes):
-                # return b"".join(d.data)
-                return bytes(d.data)
-            if isinstance(d, tlv8.Entry):
-                return tlv8.encode([d])
+                res = bytes(d.data)
+                # self.logfn(f"Transforming {d} -> {res.hex()}")
+                return res
+            elif isinstance(d, tlv8.Entry):
+                res = tlv8.encode([d])
+                # self.logfn(f"Transforming {d} -> {res.hex()}")
+                return res
             return b""
 
         encoded_structure = b"".join(map(transform, structure))
         return encoded_structure
 
     def _send_receive(
-        self, ins: Instruction, structure: Optional[List] = None
+        self,
+        ins: typing.Union[Instruction, CCIDInstruction],
+        structure: Optional[List] = None,
     ) -> bytes:
         encoded_structure = self._custom_encode(structure)
         ins_b, p1, p2 = self._encode_command(ins)
@@ -225,6 +372,7 @@ class SecretsApp:
             except Exception as e:
                 self.logfn(f"Got exception: {e}")
                 raise
+            # Data order is different here than in APDU - SW is first, then the data if any
             status_bytes, result = result[:2], result[2:]
             self.logfn(
                 f"Received [{status_bytes.hex()}] {result.hex() if result else result!r}"
@@ -240,16 +388,26 @@ class SecretsApp:
                 f"Received final data: [{status_bytes.hex()}] {data_final.hex() if data_final else data_final!r}"
             )
 
+        if data_final:
+            try:
+                self.logfn(
+                    f"Decoded received: {[e.data for e in tlv8.decode(data_final)]}"
+                )
+            except Exception:
+                pass
+
         return data_final
 
     @classmethod
-    def _encode_command(cls, command: Instruction) -> bytes:
+    def _encode_command(
+        cls, command: typing.Union[Instruction, CCIDInstruction]
+    ) -> bytes:
         p1 = 0
         p2 = 0
         if command == Instruction.Reset:
             p1 = 0xDE
             p2 = 0xAD
-        elif command == Instruction.Select:
+        elif command == CCIDInstruction.Select:
             p1 = 0x04
             p2 = 0x00
         elif command == Instruction.Calculate or command == Instruction.CalculateAll:
@@ -264,19 +422,69 @@ class SecretsApp:
         self.logfn("Executing reset")
         self._send_receive(Instruction.Reset)
 
-    def list(self) -> List[bytes]:
+    def list(
+        self, extended: bool = False
+    ) -> typing.List[typing.Union[typing.Tuple[bytes, bytes], bytes]]:
         """
         Return a list of the registered credentials
-        :return: List of bytestrings
+        :return: List of bytestrings, or tuple of bytestrings, if "extended" switch is provided
+        @deprecated
         """
         raw_res = self._send_receive(Instruction.List)
         resd: tlv8.EntryList = tlv8.decode(raw_res)
+        res: List[typing.Union[typing.Tuple[bytes, bytes], bytes]] = []
+        for e in resd:
+            # e: tlv8.Entry
+            if extended:
+                res.append((e.data[0], e.data[1:]))
+            else:
+                res.append(e.data[1:])
+        return res
+
+    def list_with_properties(self, version: int = 1) -> List[ListItem]:
+        """
+        Return a list of the registered credentials with properties
+        :return: List of ListItems
+        """
+        data = [RawBytes([version])]
+        raw_res = self._send_receive(Instruction.List, data)
+        resd: tlv8.EntryList = tlv8.decode(raw_res)
         res = []
         for e in resd:
             # e: tlv8.Entry
-            res.append(e.data[1:])
+            attribute_byte, label, properties = e.data[0], e.data[1:-1], e.data[-1]
+            res.append(
+                ListItem(
+                    kind=Kind.from_attribute_byte_type(attribute_byte),
+                    algorithm=Algorithm.Sha1,
+                    label=label,
+                    properties=ListItemProperties.from_byte(properties),
+                )
+            )
         return res
 
+    def get_credential(self, cred_id: bytes) -> PasswordSafeEntry:
+        structure = [
+            tlv8.Entry(Tag.CredentialId.value, cred_id),
+        ]
+        raw_res = self._send_receive(Instruction.GetCredential, structure=structure)
+        resd: tlv8.EntryList = tlv8.decode(raw_res)
+        res = {}
+        self.logfn("Per field dissection:")
+        for e in resd:
+            # e: tlv8.Entry
+            res[e.type_id] = e.data
+            self.logfn(f"{hex(e.type_id)} {hex(len(e.data))}  {e.data.hex()}")
+        p = PasswordSafeEntry(
+            login=res.get(Tag.PwsLogin.value),
+            password=res.get(Tag.PwsPassword.value),
+            metadata=res.get(Tag.PwsMetadata.value),
+            name=res.get(Tag.CredentialId.value),
+            properties=res.get(Tag.Properties.value),
+        )
+        p.properties = p.properties.hex().encode() if p.properties else None
+        return p
+
     def delete(self, cred_id: bytes) -> None:
         """
         Delete credential with the given id. Does not fail, if the given credential does not exist.
@@ -288,25 +496,46 @@ class SecretsApp:
         ]
         self._send_receive(Instruction.Delete, structure)
 
+    def register_yk_hmac(self, slot: int, secret: bytes) -> None:
+        """
+        Register a Yubikey-compatible challenge-response slot.
+        @param slot: challenge-response slot
+        @param secret: the secret
+        """
+        assert slot in [1, 2]
+        self.register(
+            f"HmacSlot{slot}".encode(),
+            secret,
+            kind=Kind.Hmac,
+        )
+
     def register(
         self,
         credid: bytes,
-        secret: bytes,
-        digits: int,
-        kind: Kind = Kind.Hotp,
+        secret: bytes = b"0" * 20,
+        digits: int = 6,
+        kind: Kind = Kind.NotSet,
         algo: Algorithm = Algorithm.Sha1,
         initial_counter_value: int = 0,
         touch_button_required: bool = False,
+        pin_based_encryption: bool = False,
+        login: Optional[bytes] = None,
+        password: Optional[bytes] = None,
+        metadata: Optional[bytes] = None,
     ) -> None:
         """
-        Register new OTP credential
+        Register new OTP Credential
         :param credid: Credential ID
         :param secret: The shared key
         :param digits: Digits of the produced code
         :param kind: OTP variant - HOTP or TOTP
         :param algo: The hash algorithm to use - SHA1, SHA256 or SHA512
-        :param initial_counter_value: The counter's initial value for the HOTP credential (HOTP only)
+        :param initial_counter_value: The counter's initial value for the HOTP Credential (HOTP only)
         :param touch_button_required: User Presence confirmation is required to use this Credential
+        :param pin_based_encryption: User preference for additional PIN-based encryption
+        :param login: Login field for Password Safe
+        :param password: Password field for Password Safe
+        :param metadata: Metadata field for Password Safe
         :return: None
         """
         if initial_counter_value > 0xFFFFFFFF:
@@ -317,7 +546,7 @@ class SecretsApp:
             )
 
         self.logfn(
-            f"Setting new credential: {credid!r}, {secret.hex()}, {kind}, {algo}, counter: {initial_counter_value}"
+            f"Setting new credential: {credid!r}, {kind}, {algo}, counter: {initial_counter_value}, {touch_button_required=}, {pin_based_encryption=}"
         )
 
         structure = [
@@ -326,14 +555,26 @@ class SecretsApp:
             tlv8.Entry(
                 Tag.Key.value, bytes([kind.value | algo.value, digits]) + secret
             ),
-            RawBytes([Tag.Properties.value, 0x02 if touch_button_required else 0x00]),
+            RawBytes(
+                [
+                    Tag.Properties.value,
+                    (0x02 if touch_button_required else 0x00)
+                    | (0x04 if pin_based_encryption else 0x00),
+                ]
+            ),
             tlv8.Entry(
                 Tag.InitialCounter.value, initial_counter_value.to_bytes(4, "big")
-            ),
+            )
+            if kind in [Kind.Hotp, Kind.HotpReverse]
+            else None,
+            *PasswordSafeEntry(
+                name=credid, login=login, password=password, metadata=metadata
+            ).tlv_encode(),
         ]
+        structure = list(filter(lambda x: x is not None, structure))
         self._send_receive(Instruction.Put, structure)
 
-    def calculate(self, cred_id: bytes, challenge: int) -> bytes:
+    def calculate(self, cred_id: bytes, challenge: Optional[int] = None) -> bytes:
         """
         Calculate the OTP code for the credential named `cred_id`, and with challenge `challenge`.
         :param cred_id: The name of the credential
@@ -341,6 +582,8 @@ class SecretsApp:
             Should be equal to: timestamp/period. The commonly used period value is 30.
         :return: OTP code as a byte string
         """
+        if challenge is None:
+            challenge = 0
         self.logfn(
             f"Sending calculate request for {cred_id!r} and challenge {challenge!r}"
         )
@@ -424,19 +667,11 @@ class SecretsApp:
         ]
         self._send_receive(Instruction.SetCode, structure=structure)
 
-    def authentication_required(self, stat: Optional[SelectResponse] = None) -> bool:
-        return True
-
     def validate(self, passphrase: str) -> None:
         """
         Authenticate using a passphrase
         """
         stat = self.select()
-        if not self.authentication_required(stat):
-            # Assuming this should have been checked before calling validate()
-            raise RuntimeWarning(
-                "No passphrase is set. Authentication is not required."
-            )
         if stat.algorithm != bytes([Algorithm.Sha1.value]):
             raise RuntimeError("For the authentication only SHA1 is supported")
         challenge = stat.challenge
@@ -471,7 +706,7 @@ class SecretsApp:
         """
         AID = [0xA0, 0x00, 0x00, 0x05, 0x27, 0x21, 0x01]
         structure = [RawBytes(AID)]
-        raw_res = self._send_receive(Instruction.Select, structure=structure)
+        raw_res = self._send_receive(CCIDInstruction.Select, structure=structure)
         resd: tlv8.EntryList = tlv8.decode(raw_res)
         rd = {}
         for e in resd:
@@ -489,6 +724,7 @@ class SecretsApp:
             salt=rd.get(Tag.CredentialId.value),
             challenge=rd.get(Tag.Challenge.value),
             algorithm=rd.get(Tag.Algorithm.value),
+            serial_number=rd.get(Tag.SerialNumber.value),
         )
         return r
 
@@ -528,3 +764,24 @@ class SecretsApp:
         if self.get_feature_status_cached().challenge is not None:
             return True
         return False
+
+    def feature_pws_support(self) -> bool:
+        return self._semver_equal_or_newer("4.11.0")
+
+    def protocol_v2_confirm_all_requests_with_pin(self) -> bool:
+        # 4.7.0 version requires providing PIN each request
+        return self.get_feature_status_cached().version_str() == "4.7.0"
+
+    def protocol_v3_separate_pin_and_no_pin_space(self) -> bool:
+        # 4.10.0 makes logical separation between the PIN-encrypted and non-PIN encrypted spaces, except
+        # for overwriting the credentials
+        return self._semver_equal_or_newer("4.10.0")
+
+    def is_pin_healthy(self) -> bool:
+        counter = self.select().pin_attempt_counter
+        return not (counter is None or counter == 0)
+
+    def _semver_equal_or_newer(self, required_version: str) -> bool:
+        current = semver.Version.parse(self.get_feature_status_cached().version_str())
+        semver_req_version = semver.Version.parse(required_version)
+        return current >= semver_req_version
